#include <stdlib.h>
#include <stdio.h>
#include <err.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

char *id1 = "MFS Filesystem goes here";
char *id2 = "MFS Filesystem had better STOP here";

void do_mfs (char **), do_kernel (char **);

main (int ac, char **av)
{
    if (!av[1])
	errx (1, "commands are 'mfs' or 'kernel'");
    if (!strcmp (av[1], "mfs"))
	do_mfs (av+2);
    if (!strcmp (av[1], "kernel"))
	do_kernel (av+2);
    errx (1, "commands are 'mfs' or 'kernel'");
}

/* read file; return the malloc'ed buffer holding data. Set size to file length */
char *
read_file (char *s, int *size)
{
    FILE *f;
    struct stat st;
    char *buf;
    if ((stat (s, &st) < 0) || ((f = fopen (s, "r")) == 0))
	err (1, "%s", s);
    buf = (char *) malloc (st.st_size);
    if (!buf)
	err (1, "size=%d", st.st_size);
    if (fread (buf, 1, st.st_size, f) != st.st_size)
	err (1, s);
    fclose (f);
    if (size)
	*size = st.st_size;
    return buf;
}
/* locate the position of MFS area in s */
char *
read_kernel (char *s, int *size, int *pos, int *sectors)
{
    int i, l;
    char *buf;

    buf = read_file (s, size);
    l = strlen (id1);
    for (i = 0; i < *size; i++) 
	if (buf[i] == id1[0] && !strncmp (buf+i, id1, l))
	    break;
    if (i == *size)
	errx (1, "MFS filesystem signature not found in %s", s);
    *pos = i;

    l = strlen (id2);
    for (; i < *size; i++) 
	if (buf[i] == id2[0] && !strncmp (buf+i, id2, l))
	    break;
    if (i == *size)
	errx (1, "MFS filesystem end signature not found in %s", s);
    *sectors = (i - *pos)/512;
    return buf;
}
void
do_mfs (char **a)
{
    int pos, size, i;
    char *buf, *p, *q;
    long  s;
    char *db[2] = {"/etc/disktab", 0};

    if (!a[0])
	errx (1, "mfs requires 1 argument");

    buf = read_kernel (a[0], &i, &pos, &size);

    i = cgetfirst (&p, db);
    while (i) {
	if (cgetstr (p, "ty", &q) > 0) {
	    if (!strcmp (q, "floppy") && !cgetnum (p, "pa", &s) &&
		(s == size)) {
		q = strchr (p, ':');
		*q = 0;
		printf ("%d %s\n", size/2, p);
		exit (0);
	    }
	    free (q);
	}
	free (p);
	i = cgetnext (&p, db);
    }
    errx (1, "no suitable entry in /etc/disktab.\n"
	  "make a floppy entry with pa#%d", size);
}
void
do_kernel (char **a)
{
    int pos, size, l, sectors;
    char *buf, *m;
    FILE *f;
    struct stat st;

    if (!a[0] || !a[1] || !a[2])
	errx (1, "kernel command requires 3 arguments");
    buf = read_kernel (a[0], &size, &pos, &sectors);

    m = read_file (a[2], &l);
    if (l > sectors*512)
	errx (1, "MFS file system too small: file size=%d\n", l);
    f = fopen (a[1], "w");
    if (!f)
	err (1, a[1]);
    if (fwrite (buf, 1, pos+8192, f) != pos+8192)
	err (1, a[1]);
    if (fwrite (m+8192, 1, l-8192, f) != l-8192)
	err (1, a[1]);
    if (fwrite (buf+pos+sectors*512, 1, size-(pos+sectors*512), f) != 
	size-(pos+sectors*512))
	err (1, a[1]);
    exit (0);
}
